#############################################################################
#
#	configuration.py
#	Sets up Pyro's configuration (Pyro.config).
#
#	This is part of "Pyro" - Python Remote Objects
#	Which is (c) Irmen de Jong - irmen@users.sourceforge.net
#
#############################################################################


# Initialize Pyro Configuration.

import re, os
import util
from errors import PyroError



# ---------------------- DEFAULT CONFIGURATION VARIABLES -----------

# Special characters are '%c' (current directory, absolute) and
# $STORAGE which is replaced by the PYRO_STORAGE path.
_defaults= {
	# 'PYRO_CONFIG_FILE': this one has no default
	'PYRO_STORAGE':			'%c',	# current dir (abs)
	'PYRO_PORT':			7766,
	'PYRO_PORT_RANGE':		100,
	'PYRO_NS_PORT':			9090,
	'PYRO_NS_BC_PORT':		9091,
	'PYRO_NS_URIFILE':		'$STORAGE/Pyro_NS_URI', # (abs)
	'PYRO_NS_DEFAULTGROUP': ':Default',
	'PYRO_NS_HOSTNAME':		None,
	'PYRO_BC_RETRIES':		2,
	'PYRO_BC_TIMEOUT':		2,
	'PYRO_BINARY_PICKLE':	1,
	'PYRO_XML_PICKLE':		None,
	'PYRO_TRACELEVEL':		0,
	'PYRO_USER_TRACELEVEL':	0,
	'PYRO_LOGFILE':			'$STORAGE/Pyro_log',		# (abs)
	'PYRO_USER_LOGFILE':	'$STORAGE/Pyro_userlog',	# (abs)
	'PYRO_MAXCONNECTIONS':	200,
	'PYRO_MULTITHREADED':	util.supports_multithreading(),
	'PYRO_COMPRESSION':		0,
	'PYRO_MOBILE_CODE':		0,
	'PYRO_DNS_URI':			0,
	'PYRO_CHECKSUM':		0,
	'PYRO_ES_QUEUESIZE':	1000,
	'PYRO_ES_BLOCKQUEUE':	1,
	'PYROSSL_CERTDIR':		'$STORAGE/certs',
	'PYROSSL_CA_CERT':		'ca.pem',
	'PYROSSL_SERVER_CERT':	'server.pem',
	'PYROSSL_CLIENT_CERT':	'client.pem',
}

# ---------------------- END OF DEFAULT CONFIGURATION VARIABLES -----


class Config:

	def __init__(self):
		pass

	def setup(self, configFile):
		global _defaults
		reader = ConfigReader(_defaults)
		try:
			reader.parse(configFile)
		except EnvironmentError,x:
			raise PyroError("Error reading config file: "+configFile+"; "+str(x));
		self.__dict__.update(reader.items)
		if(configFile):
			self.__dict__['PYRO_CONFIG_FILE'] = os.path.abspath(configFile)
		else:
			self.__dict__['PYRO_CONFIG_FILE'] = ''


		# Create the storage directory if it doesn't exist yet
		if not os.path.exists(self.PYRO_STORAGE):
			os.mkdir(self.PYRO_STORAGE)
		# see if we have permission there
		if not os.path.isdir(self.PYRO_STORAGE):
			raise IOError('PYRO_STORAGE is not a directory ['+self.PYRO_STORAGE+']')
		tstfile=os.path.join(self.PYRO_STORAGE,'access_test_file')
		try:
			open(tstfile,'w')
			os.remove(tstfile)
		except:
			raise IOError('no write access to PYRO_STORAGE ['+self.PYRO_STORAGE+']')

#	def __getattr__(self,name):
#		# add smart code here to deal with other requested config items!



class ConfigReader:
	def __init__(self, defaults):
		self.matcher=re.compile(r'\s*([^#]\w*)\s*=\s*(\S*)\s*$')
		self.items=defaults.copy()

	def parse(self, file):
		global _defaults
		if file:
			file=open(file).readlines()
			for l in file:
				match=self.matcher.match(l)
				if match:
					if match.group(1) in _defaults:
						if match.group(2):
							self.items[match.group(1)] = match.group(2)
					else:
						raise KeyError('Unknown config item in configfile: '+match.group(1))

		# Parse the environment variables (they override the config file)
		self.items.update(self.processEnv(_defaults.keys()))

		# First, fix up PYRO_STORAGE because others depend on it.
		self.items['PYRO_STORAGE'] = self.treatSpecial(self.items['PYRO_STORAGE'])
		# Now fix up all other items:
		for i in self.items.keys():
			newVal = self.treatSpecial(self.items[i])
			if i in ('PYRO_STORAGE', 'PYRO_LOGFILE', 'PYRO_USER_LOGFILE', 'PYRO_NS_URIFILE'):
				newVal=os.path.abspath(newVal)
			# fix the variable type if it's an integer
			if type(_defaults[i]) == type(42):
				newVal = int(newVal)
			self.items[i]= newVal

	def processEnv(self, keys):
		env={}
		for key in keys:
			try: env[key] = os.environ[key]
			except KeyError: pass
		return env

	def treatSpecial(self, value):
		# treat special escape strings
		if type(value)==type(""):
			if value=='%c':
				return os.curdir
			elif len(value)>=9 and value[:9]=='$STORAGE/':
				return os.path.join(self.items['PYRO_STORAGE'], value[9:])
		return value


