#############################################################################
#
#	util.py - Pyro Utilities
#
#	This is part of "Pyro" - Python Remote Objects
#	which is (c) Irmen de Jong - irmen@users.sourceforge.net
#
#############################################################################


import os,random,time,string,sys,binascii
import Pyro


def supports_multithreading():
	try:
		import threading; return 1
	except:
		return 0
	
def supports_compression():
	try:
		import zlib; return 1
	except:
		return 0

if supports_multithreading():
	import threading


# bogus lock class, for systems that don't have threads.
class BogusLock:
	def acquire(self): pass
	def release(self): pass

def getLockObject():	
	if supports_multithreading():
		from threading import Semaphore
		return Semaphore()
	else:
		return BogusLock()

# Log classes for (M)essages, (W)arnings, (E)rrors.


class LoggerBase:
	# Logger base class. Subclasses must override _logfile and  _checkTraceLevel.
	def __init__(self):
		import Pyro
		self.lock=getLockObject()
	def msg(self,source,*args):
		if self._checkTraceLevel(3): self._trace('NOTE',source, args)
	def warn(self,source,*args):
		if self._checkTraceLevel(2): self._trace('WARN',source, args)
	def error(self,source,*args):
		if self._checkTraceLevel(1): self._trace('ERR!',source, args)
	def raw(self,str):
		try:
			self.lock.acquire()
			f=open(self._logfile(),'a')
			f.write(str)
			f.close()
			self.lock.release()
		except:
			pass
	def _trace(self,typ,source, arglist):
		self.lock.acquire()
		if not arglist:
			arglist=['<no message>']
		try:
			tf=open(self._logfile(),'a')
			try:
				pid=os.getpid()
				pidinfo=" ["+str(os.getpid())
			except:
				pidinfo=" ["   # XXX jython has no getpid()
			if supports_multithreading():
				pidinfo+=":"+threading.currentThread().getName()
			pidinfo+="] "	
			tf.write(time.strftime('%Y-%m-%d %H:%M:%S',time.localtime(time.time()))+
			  pidinfo+'** '+typ+' ** '+str(source)+' ** '+reduce(lambda x,y: str(x)+' '+str(y),arglist)+'\n')
			tf.close()
		except Exception,x:
			pass
		self.lock.release()
	def _logfile(self):
		raise NotImplementedError,'must override'
	def _checkTraceLevel(self,level):
		raise NotImplementedError,'must override'

class SystemLogger(LoggerBase):
	def _checkTraceLevel(self, level):
		return Pyro.config.PYRO_TRACELEVEL >= level
	def _logfile(self):
		file=Pyro.config.PYRO_LOGFILE
		if not os.path.isabs(file):
			Pyro.config.PYRO_LOGFILE=os.path.join(Pyro.config.PYRO_STORAGE, file)
		return Pyro.config.PYRO_LOGFILE
		
class UserLogger(LoggerBase):
	def _checkTraceLevel(self, level):
		return Pyro.config.PYRO_USER_TRACELEVEL >= level
	def _logfile(self):
		file=Pyro.config.PYRO_USER_LOGFILE
		if not os.path.isabs(file):
			Pyro.config.PYRO_USER_LOGFILE=os.path.join(Pyro.config.PYRO_STORAGE, file)
		return Pyro.config.PYRO_USER_LOGFILE


# The logger object 'Log'.
Log = SystemLogger()


# Caching directory lister, outputs (filelist,dirlist) tuple
# Based upon dircache.py, but implemented in a callable object
# that has a thread-safe cache.
class DirLister:
	def __init__(self):
		self.lock=getLockObject()
		self.__listdir_cache = {}

	def __call__(self,path):
		self.lock.acquire()
		try:
			cached_mtime, files, directories = self.__listdir_cache[path]
			del self.__listdir_cache[path]
		except KeyError:
			cached_mtime, files, directories = -1, [], []
		self.lock.release()
		mtime = os.stat(path)[8]
		if mtime <> cached_mtime:
			list = os.listdir(path)
			files=[]
			directories=[]
			for e in list:
				if os.path.isdir(os.path.join(path,e)):
					directories.append(e)
				else:
					files.append(e)
		self.lock.acquire()			
		self.__listdir_cache[path] = mtime, files, directories
		self.lock.release()
		return files,directories

listdir = DirLister()		# callable object		


# Fairly simple argument options parser. Like getopt(3).
class ArgParser:
	def __init__(self):
		pass
	def parse(self, args, optionlist):
		# optionlist is a string such as "ab:c" which means
		# we search for 3 options (-a, -b, -c) of which -b has an argument.
		self.options={}		# public, the option->value dictionary
		self.args=[]		# public, the rest of the arguments
		self.ignored=[]		# public, ignored options
		optionlist+=' '		# add sentinel
		if type(args)==type(''):
			args=string.split(args)
		while args:
			arg=args[0]
			del args[0]
			if arg[0]=='-':
				if len(arg)>=2:   # arg is an option. Check our list
					idx = string.find(optionlist,arg[1])
					if idx>=0:
						if optionlist[idx+1]==':':   # option requires argument.
							if len(arg)>=3:   # argument is appended. Use this.
								self.options[arg[1]]=arg[2:]
								continue
							# fetch argument from next string
							if len(args)>=1:
								self.options[arg[1]]=args[0]
								del args[0]
								continue
							else:   # missing arg, substitute None
								self.options[arg[1]]=None
						else:   # option requires no argument, use None
							self.options[arg[1]]=None
					else:   # didn't find this option, skip it
						self.ignored.append(arg[1])
				else:   # arg is a single '-'. Stop parsing.
					for a in args:
						self.args.append(a)
					args=None
			else:   # arg is no option, add it to the residu list and continue
				self.args.append(arg)
	def hasOpt(self, option):
		return option in self.options
	def getOpt(self, option, default=Exception()):
		try:
			return self.options[option]
		except KeyError:
			if not isinstance(default,Exception):
				return default
			raise KeyError('no such option')
	def printIgnored(self):
		if self.ignored:
			print 'Ignored options:',
			for o in self.ignored:
				print '-'+o,
			print


def getGUID():
	# Generate readable GUID string.
	# The GUID is constructed as follows: hexlified string of
	# AAAAAAAA-AAAABBBB-BBBBBBBB-BBCCCCCC  (a 128-bit number in hex)
	# where A=network address, B=timestamp, C=random. 
	# The 128 bit number is returned as a string of 16 8-bits characters.
	# For A: should use the machine's MAC ethernet address, but there is no
	# portable way to get it... use the IP address + 2 bytes process id.

	from Pyro.protocol import getIPAddress
	
	ip=getIPAddress()
	if ip:
		ip=map(int, string.split(ip,'.'))
		ip=reduce(lambda x,y: (x<<8)+y, ip, 0L)
	else:
		# can't get IP address... use another value, like our Python id() and PID
		Log.warn('getGUID','Can\'t get IP address')
		try:
			ip=os.getpid()
		except:
			ip=0
		ip += id(getGUID)

	if os.name!='java':
		extra=os.getpid()
	else:
		extra=id(os)		# XXX preliminary Jython compat. (no getpid)

	networkAddr = (long(ip) << 16 ) | extra
	t1=time.time()*100
	t2=int((t1*time.clock())%sys.maxint) & 0xffffff
	t1=int(t1%sys.maxint) 
	timestamp = (long(t1) << 24) | t2
	r2=(random.randint(0,sys.maxint/2)>>4) & 0xffff
	r3=(random.randint(0,sys.maxint/2)>>5) & 0xff
	rndm = (r2<<8) | r3
	# return hexlified string: (see also binascii.hexlify)
	return '%012x%014x%06x' % (networkAddr & 0xffffffffffffl,
	                             timestamp & 0xffffffffffffffl,
	                                  rndm & 0xffffff)

def genguid_scripthelper():
	if __name__!='__main__':
		del sys.argv[0]	# remove '-c' arg
	p=ArgParser()
	p.parse(sys.argv,'')
	if p.args or p.ignored:
		print 'usage: genguid  (no arguments)'
		raise SystemExit
	print getGUID()



# create a weak reference proxy if available (Python 2.1+)
def weakref(obj):
	if obj is None:
		return obj
	try:
		import weakref
		return weakref.proxy(obj)
	except ImportError:	
		return obj


# Get the configured pickling module. 
# Currently supported: cPickle, pickle, gnosis.xml.pickle (@paranoia -1),
# and PyXML 0.8+'s xml.marshal.generic.
def getPickle():
	if Pyro.config.PYRO_XML_PICKLE:
		# user requires xml pickle. Fails if that is not available!
		return getXMLPickle()
	else:	
		try: 
			import cPickle
			return cPickle
		except ImportError:
			# Fall back on pickle if cPickle isn't available
			import pickle
			return pickle

_xmlpickle={}
# the following wrapper is necessary because PyXML's pickler
# cannot take a 'binary' third argument...
class PyXMLPickler:
	def __init__(self):
		import xml.marshal.generic
		self.pickle=xml.marshal.generic
	def dumps(self, obj, bin=None):
		return self.pickle.dumps(obj)
	def dump(self, obj, file, bin=None):
		return self.pickle.dumps(obj,file)
	def loads(self, str):
		return self.pickle.loads(str)
	def load(self, file):
		return self.pickle.load(file)
	
def getXMLPickle(impl=None):		
	# load & config the required xml pickle.
	# Currently supported: Gnosis Utils' gnosis.xml.pickle,
	# and PyXML 0.8+'s xml.marshal.generic
	global _xmlpickle
	if not impl:
		impl=Pyro.config.PYRO_XML_PICKLE
	impl=impl.lower()
	if impl=='any':
		# first try if PyXML is available, otherwise, choose Gnosis.
		impl='pyxml'
		Pyro.config.PYRO_XML_PICKLE=impl
		try:
			import xml.marshal.generic
		except ImportError:
			impl='gnosis'
			Pyro.config.PYRO_XML_PICKLE=impl
	if impl in _xmlpickle:
		return _xmlpickle[impl]
	try:	
		if impl=='pyxml': # PyXML init
			_xmlpickle['pyxml']=PyXMLPickler()
			return _xmlpickle['pyxml']
		elif impl=='gnosis': # Gnosis init
			import gnosis.xml.pickle
			_xmlpickle['gnosis']=gnosis.xml.pickle
			gnosis.xml.pickle.setParanoia(0)		# default paranoia level is too strict for Pyro
			gnosis.xml.pickle.setParser('SAX')		# use fastest parser (cEXPAT?)
			return gnosis.xml.pickle
		else:
			raise ImportError('unsupported xml pickle implementation requested: '+impl)
	except ImportError:
		Log.error('server was asked to use xml pickling but implementation ('+impl+') is not available')
		raise NotImplementedError('server was asked to use xml pickling but implementation ('+impl+') is not available')


# Pyro traceback printing
def getPyroTraceback(exc_obj):
	import traceback, constants, sys
	try:
		exc_type, exc_value, exc_trb=sys.exc_info()
		remote_tb=getattr(exc_obj,constants.TRACEBACK_ATTRIBUTE,None)
		l_tb_list=traceback.format_exception(exc_type, exc_value, exc_trb)
		local_tb=traceback.format_exception(exc_type, exc_value, exc_trb)
		if remote_tb:
			return [ '---- remote Pyro traceback ----\n' ] + \
				remote_tb + [ '---- local traceback ----\n' ] +  local_tb
		else:
			# hmm. no remote tb info, return just the local tb.
			return local_tb
	finally:
		# clean up cycle to traceback
		del exc_type, exc_value, exc_trb


